import logging
import time
from collections.abc import Sequence
from typing import Any

import urllib3
from selenium.common.exceptions import TimeoutException, WebDriverException
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from promium._diag import push_step
from promium.exceptions import PromiumException, PromiumTimeout
from promium.logger import find_console_browser_errors, repr_console_errors


log = logging.getLogger(__name__)

NORMAL_LOAD_TIME = 10

http = urllib3.PoolManager(
    cert_reqs=False,
    timeout=10,
    headers={
        "Accept-Encoding": "gzip, deflate",
        "Accept": "*/*",
        "Connection": "keep-alive",
    },
)


def enable_jquery(driver: Any) -> None:
    """Inject jQuery into the page if it is not already loaded."""
    driver.execute_script(
        """
        jqueryUrl =
        'https://ajax.googleapis.com/ajax/libs/jquery/3.2.1/jquery.min.js';
        if (typeof jQuery == 'undefined') {
            var script = document.createElement('script');
            var head = document.getElementsByTagName('head')[0];
            var done = false;
            script.onload = script.onreadystatechange = (function() {
                if (!done && (!this.readyState || this.readyState == 'loaded'
                        || this.readyState == 'complete')) {
                    done = true;
                    script.onload = script.onreadystatechange = null;
                    head.removeChild(script);

                }
            });
            script.src = jqueryUrl;
            head.appendChild(script);
            console.log("injection Complete by script enable jquery")
        };
        """
    )
    wait_obj = WebDriverWait(driver, timeout=NORMAL_LOAD_TIME)
    wait_obj.until(
        lambda d: d.execute_script("return window.jQuery != undefined") is True,
        message="Jquery script injection - failed ",
    )


def _check_browser_console(driver: Any) -> None:
    """Check browser console errors when exception occurs."""
    if console_errors := find_console_browser_errors(driver):
        log.warning(repr_console_errors(console_errors))


def wait(
    driver: Any,
    seconds: float = NORMAL_LOAD_TIME,
    poll_frequency: float = 0.5,
    ignored_exceptions: Sequence[type[BaseException]] | None = None,
) -> WebDriverWait:
    """Return a configured WebDriverWait object."""
    return WebDriverWait(
        driver=driver,
        timeout=seconds,
        poll_frequency=poll_frequency,
        ignored_exceptions=ignored_exceptions or [WebDriverException],
    )


def wait_until(
    driver: Any,
    expression: Any,
    seconds: float = NORMAL_LOAD_TIME,
    msg: str | None = None,
) -> Any:
    """Wait until expression is True or timeout expires."""
    try:
        try:
            name = getattr(expression, "__name__", "<expr>")
            push_step(f'wait_until("{name}", {seconds}s)')
        except Exception:
            pass
        return wait(driver, seconds).until(expression, msg)
    except TimeoutException as e:
        _check_browser_console(driver)
        raise PromiumTimeout(e.msg, seconds, e.screen) from e


def wait_until_not(
    driver: Any,
    expression: Any,
    seconds: float = NORMAL_LOAD_TIME,
    msg: str | None = None,
) -> Any:
    """Wait until expression becomes False or timeout expires."""
    try:
        return wait(driver, seconds).until_not(expression, msg)
    except TimeoutException as e:
        _check_browser_console(driver)
        raise PromiumTimeout(e.msg, seconds, e.screen) from e


def wait_until_with_reload(
    driver: Any,
    expression: Any,
    trying: int = 5,
    seconds: float = 2,
    msg: str = "",
) -> None:
    """Wait until expression is True, refreshing page on each attempt."""
    for t in range(trying):
        try:
            if wait_until(driver, expression, seconds=seconds):
                return
        except PromiumException:
            log.info(f"Waiting `until` attempt number {t + 1}.")
            driver.refresh()
    msg = f"\n{msg}" if msg else ""
    raise PromiumTimeout(
        f"The values in expression not return true after {trying} tries {msg}",
        seconds=trying * seconds,
    )


def wait_until_not_with_reload(
    driver: Any,
    expression: Any,
    trying: int = 5,
    seconds: float = 2,
    msg: str = "",
) -> None:
    """Wait until expression is False, refreshing page on each attempt."""
    for t in range(trying):
        try:
            if not wait_until_not(driver, expression, seconds=seconds):
                return
        except PromiumException:
            log.info(f"Waiting `until not` attempt number {t + 1}.")
            driver.refresh()
    msg = f"\n{msg}" if msg else ""
    raise PromiumTimeout(
        f"The values in expression not return false after {trying} tries {msg}.",
        seconds=trying * seconds,
    )


def wait_for_animation(driver: Any) -> Any:
    """Wait until all jQuery animations are finished."""
    jquery_script = 'return jQuery(":animated").length == 0;'
    return wait_until(
        driver=driver,
        expression=lambda d: d.execute_script(jquery_script),
        seconds=NORMAL_LOAD_TIME,
        msg=f"Animation timeout (waiting time: {NORMAL_LOAD_TIME} sec)",
    )


def wait_document_status(driver: Any, seconds: float = NORMAL_LOAD_TIME) -> None:
    """Wait until document.readyState becomes 'complete'."""
    wait_obj = WebDriverWait(driver, timeout=seconds)
    try:
        wait_obj.until(
            lambda d: d.execute_script("return document.readyState")
            == "complete",
            message=(
                "Page didn't load because document.readyState did not go "
                "into status 'complete'"
            ),
        )
    except TimeoutException as e:
        _check_browser_console(driver)
        raise PromiumTimeout(e.msg, seconds, e.screen) from e


def wait_websocket_connection_open(
    driver: Any, seconds: float = NORMAL_LOAD_TIME
) -> None:
    """Wait until WebSocket connection is open."""
    wait_obj = WebDriverWait(driver, timeout=seconds)
    try:
        wait_obj.until(
            lambda d: d.execute_script("return WebSocket.OPEN") == 1,
            message="ReadyState of WebSocket connection didn't change on OPEN",
        )
    except TimeoutException as e:
        _check_browser_console(driver)
        raise PromiumTimeout(e.msg, seconds, e.screen) from e


def wait_jquery_status(driver: Any, seconds: float = NORMAL_LOAD_TIME) -> None:
    """Wait until jQuery active requests count becomes 0."""
    wait_obj = WebDriverWait(driver, timeout=seconds)
    try:
        wait_obj.until(
            lambda d: d.execute_script("return jQuery.active") == 0,
            message=(
                "Page didn't load because jquery script did not go into "
                "'active == 0' status"
            ),
        )
    except TimeoutException as e:
        _check_browser_console(driver)
        raise PromiumTimeout(e.msg, seconds, e.screen) from e


def wait_for_page_loaded(driver: Any, seconds: float = NORMAL_LOAD_TIME) -> None:
    """Wait until page is fully loaded with document, jQuery, and no alert."""
    wait_document_status(driver, seconds)
    enable_jquery(driver)
    wait_jquery_status(driver, seconds)
    alert_is_present = EC.alert_is_present()
    if alert_is_present(driver):
        alert_text = driver.switch_to_alert().text
        driver.switch_to_alert().dismiss()
        raise alert_text


def wait_for_alert(driver: Any, seconds: float = NORMAL_LOAD_TIME) -> Any:
    """Wait for alert to be present."""
    return WebDriverWait(
        driver, seconds, ignored_exceptions=[WebDriverException]
    ).until(EC.alert_is_present(), "Alert is not present.")


def wait_for_alert_is_displayed(
    driver: Any, seconds: float = NORMAL_LOAD_TIME
) -> bool:
    """Return True if alert is displayed within timeout, else False."""
    try:
        wait_for_alert(driver, seconds=seconds)
    except TimeoutException:
        return False
    return True


def wait_for_status_code(
    url: str,
    status_code: int = 200,
    tries: int = 10,
    allow_redirects: bool = True,
) -> Any:
    """Wait until given URL returns expected status code."""
    current_status: int | None = None
    for _ in range(tries):
        response = http.request(
            url=url,
            method="GET",
            redirect=allow_redirects,
        )
        current_status = response.status
        if current_status == status_code:
            return response
        time.sleep(1)
    raise PromiumException(
        f"Url {url} return status code {current_status} after {tries} tries."
    )


def wait_until_new_window_is_opened(
    driver: Any, original_window: str, window_num: int = 0
) -> str:
    """Wait until a new browser window is opened and return its handle."""
    wait_until(
        driver,
        lambda x: len(driver.window_handles) >= 2,
        msg="New window was not opened.",
    )
    window_handles = driver.window_handles
    window_handles.remove(original_window)
    return window_handles[window_num]


def wait_part_appear_in_class(
    driver: Any, obj: Any, part_class: str, msg: str | None = None
) -> None:
    """Wait until given substring appears in element's class attribute."""
    if not msg:
        msg = f'"{part_class}" not present in attribute *class*.'
    wait_until(
        driver,
        lambda d: (part_class in obj.get_attribute("class")),
        seconds=5,
        msg=msg,
    )


def wait_part_disappear_in_class(
    driver: Any, obj: Any, part_class: str, msg: str | None = None
) -> None:
    """Wait until given substring disappears from element's class attribute."""
    if not msg:
        msg = f'"{part_class}" not present in attribute *class*.'
    wait_until(
        driver,
        lambda d: (part_class not in obj.get_attribute("class")),
        seconds=5,
        msg=msg,
    )


def wait_url_contains(
    driver: Any, text: str, msg: str | None = None, timeout: int = 5
) -> None:
    """Wait until current URL contains given text."""
    if not msg:
        msg = f'"{text}" not present in current url {driver.current_url}.'
    wait_until(driver, EC.url_contains(text), seconds=timeout, msg=msg)


def wait_url_not_contains(
    driver: Any, text: str, msg: str | None = None, timeout: int = 5
) -> None:
    """Wait until current URL does not contain given text."""
    if not msg:
        msg = f'"{text}" not disappear in url.'
    wait_until_not(driver, EC.url_contains(text), seconds=timeout, msg=msg)



import os

import pytest
from selenium.webdriver.firefox.options import Options as FirefoxOptions

from promium.device_config import CustomDevice
from promium.support.settings import SettingsFirefox


@pytest.mark.unit
def test_driver_capabilities():
    assert SettingsFirefox.driver_capabilities
    assert isinstance(SettingsFirefox.driver_capabilities, dict)
    assert SettingsFirefox.driver_capabilities.get("browserName") == "firefox"


@pytest.mark.unit
def test_binary_path():
    assert SettingsFirefox.binary_path
    assert "gekodriver" in SettingsFirefox.binary_path


@pytest.mark.unit
def test_default_options():
    assert len(SettingsFirefox.default_options) > 0
    for opt in SettingsFirefox.default_options:
        assert isinstance(opt, str)


@pytest.mark.unit
def test_default_preferences():
    assert len(SettingsFirefox.default_preferences) > 0
    for opt in SettingsFirefox.default_options:
        assert isinstance(opt, str)


@pytest.mark.unit
def test_driver_options():
    assert isinstance(SettingsFirefox.driver_options(), FirefoxOptions)


@pytest.mark.unit
def test_set_windows_size():
    SettingsFirefox.device = CustomDevice(
        width=123, height=321, user_agent="xxx"
    )
    SettingsFirefox.set_windows_size()
    ex_options = [
        f"-width {SettingsFirefox.device.width}",
        f"-height {SettingsFirefox.device.height}",
    ]
    for opt in ex_options:
        assert opt in SettingsFirefox.default_options


@pytest.mark.unit
def test_set_windows_size_with_device():
    device = CustomDevice(width=123, height=321, user_agent="user xxx")
    SettingsFirefox.set_windows_size(device)
    ex_options = [
        f"-width {SettingsFirefox.device.width}",
        f"-height {SettingsFirefox.device.height}",
    ]
    for opt in ex_options:
        assert opt in SettingsFirefox.default_options


@pytest.mark.unit
def test_get_options_with_default_values():
    os.environ["HEADLESS"] = ""
    SettingsFirefox.user_exp_options = {}
    SettingsFirefox.user_options = []
    opt = SettingsFirefox.get_options()
    assert isinstance(opt, FirefoxOptions)
    assert SettingsFirefox.default_options == opt.arguments


@pytest.mark.unit
def test_get_options_with_user_values():
    os.environ["HEADLESS"] = ""
    SettingsFirefox.user_exp_options = {"key1": 1}
    SettingsFirefox.user_options = ["param1", "param2", "--param3", "-p4"]
    opt = SettingsFirefox.get_options()
    assert isinstance(opt, FirefoxOptions)
    assert SettingsFirefox.user_options == opt.arguments


@pytest.mark.unit
def test_get_options_without_headless():
    os.environ["HEADLESS"] = "false"
    SettingsFirefox.user_exp_options = {}
    SettingsFirefox.user_options = []
    opt = SettingsFirefox.get_options()
    assert isinstance(opt, FirefoxOptions)
    assert "-headless" not in opt.arguments


@pytest.mark.unit
def test_get_options_with_headless():
    os.environ["HEADLESS"] = "Enabled"
    SettingsFirefox.user_exp_options = {}
    SettingsFirefox.user_options = []
    opt = SettingsFirefox.get_options()
    assert isinstance(opt, FirefoxOptions)
    assert "-headless" in opt.arguments


@pytest.mark.unit
def test_get_preferences():
    profile = SettingsFirefox.get_preferences()
    for pr in SettingsFirefox.default_preferences.keys():
        assert pr in profile._desired_preferences


@pytest.mark.unit
def test_get_preferences_with_user_values():
    SettingsFirefox.user_preferences = {"key1": 1}
    profile = SettingsFirefox.get_preferences()
    for user_pr in SettingsFirefox.user_preferences.keys():
        assert user_pr in profile._desired_preferences


@pytest.mark.unit
@pytest.mark.parametrize("user_agent", ["", None])
def test_get_preferences_with_user_agent_preferences(user_agent):
    SettingsFirefox.device = CustomDevice(
        width=123, height=321, user_agent=user_agent
    )
    profile = SettingsFirefox.get_preferences()
    assert not profile._desired_preferences.get("general.useragent.override")


@pytest.mark.unit
def test_get_preferences_with_user_agent():
    SettingsFirefox.device = CustomDevice(
        width=123, height=321, user_agent="user agent xxx"
    )
    profile = SettingsFirefox.get_preferences()
    user_agent = profile._desired_preferences.get("general.useragent.override")
    assert user_agent == "user agent xxx"



import inspect
import json
import logging
import re
from deepdiff import DeepDiff
from pathlib import Path
from typing import Any, TypeVar

import allure
import urllib3
from json_checker import Checker, CheckerError
from selenium.webdriver.chrome.webdriver import WebDriver

from promium.common import upload_screenshot


CoerceDict = dict[str, type] | None

log = logging.getLogger(__name__)

http = urllib3.PoolManager(
    cert_reqs=False,
    timeout=5,
    headers={
        "Accept-Encoding": "gzip, deflate",
        "Accept": "*/*",
        "Connection": "keep-alive",
    },
)


def base_msg(msg: str) -> str:
    """Return message with trailing newline if provided."""
    return f"{msg}\n" if msg else ""


T = TypeVar("T")


def _check_namedtuple[T](obj: T) -> dict[str, Any] | T:
    """Convert namedtuple to dict if possible, otherwise return the object."""
    if hasattr(obj, "_asdict"):
        return obj._asdict()
    return obj


def convert_container(container: Any) -> dict | str:
    """Convert a container to a formatted JSON string, if not already
    a string."""
    if not isinstance(container, str):
        return json.dumps(
            obj=container,
            indent=4,
            sort_keys=True,
            ensure_ascii=False,
            default=str,
        )
    return _check_namedtuple(container)


def get_text_with_ignore_whitespace_symbols(text: str) -> str:
    """Return text with all whitespace symbols removed."""
    text_without_whitespace_symbols = (
        text.replace("\t", " ")
        .replace("\v", " ")
        .replace("\r", " ")
        .replace("\n", " ")
        .replace("\f", " ")
        .strip()
    )
    text_list = text_without_whitespace_symbols.split(" ")
    text_list_without_space = [word for word in text_list if word]
    return " ".join(text_list_without_space)


class BaseSoftAssertion:
    """Base class for soft assertion utilities."""

    # TODO is not cleaned in unit tests need use __init__
    assertion_errors = []

    def _relpath(self, p: str) -> str:
        """Return relative path from current working directory."""
        try:
            return Path(p).resolve().relative_to(Path.cwd()).as_posix()
        except Exception:
            return Path(p).as_posix()

    def _callsite_line(self) -> str:
        """Return callsite line in format ↳ at path:lineno."""
        for frame in inspect.stack()[2:]:
            fn = frame.filename.replace("\\", "/")
            if "/site-packages/" in fn:
                continue
            return f"↳ at {self._relpath(fn)}:{frame.lineno}"
        f = inspect.stack()[2]
        return f"↳ at {self._relpath(f.filename)}:{f.lineno}"

    def add_screenshot(self, error_msg: str = "soft_assertion") -> str:
        """Attach screenshot to allure report if driver is available."""
        if hasattr(self, "driver"):
            try:
                allure.attach(
                    self.driver.get_screenshot_as_png(),
                    name=error_msg,
                    attachment_type=allure.attachment_type.PNG,
                )
            except Exception as e:
                log.error(f"[PROMIUM] Can't attach screenshot: {e}")
            return f"\nScreenshot: {upload_screenshot(self.driver)}\n"
        return ""

    def soft_assert_true(
        self, expr: Any, msg: str = None, with_screen: bool = False
    ) -> str | None:
        """Check that the expression is true."""
        if not expr:
            error = msg or "Is not true."
            screen = self.add_screenshot(error) if with_screen else ""
            self.assertion_errors.append(
                f"{error}{screen}\n{self._callsite_line()}\n"
            )
            return error

    def soft_assert_false(
        self, expr: Any, msg: str = None, with_screen: bool = False
    ) -> str | None:
        """Check that the expression is false."""
        if expr:
            error = msg or "Is not false."
            screen = self.add_screenshot(error) if with_screen else ""
            self.assertion_errors.append(
                f"{error}{screen}\n{self._callsite_line()}\n"
            )
            return error

    def soft_assert_equals(
        self,
        current: Any,
        expected: Any,
        msg: str = None,
        show_diff: bool = False,
        with_screen: bool = False,
    ) -> str | None:
        """Check that current equals expected, optionally showing diff."""
        diff = None
        if type(current) is list and type(expected) is list:
            diff = DeepDiff(current, expected, ignore_order=True)
        if show_diff and type(current) is str and type(expected) is str:
            current_splitlines = current.splitlines()
            expected_splitlines = expected.splitlines()
            if len(current_splitlines) > 1 or len(expected_splitlines) > 1:
                diff = DeepDiff(current_splitlines, expected_splitlines)
        difference = f"\nDifference:\n{convert_container(diff)}" if diff else ""

        type_current = type_expected = ""
        if type(current) is not type(expected):
            type_current = f"({type(current)})"
            type_expected = f"({type(expected)})"

        assert_message = (
            f"{base_msg(msg)}\n"
            f"Current - {convert_container(current)} {type_current}\n"
            f"Expected - {convert_container(expected)} {type_expected}\n"
            f"{difference}"
        )
        return self.soft_assert_true(
            current == expected, assert_message, with_screen
        )

    def soft_assert_not_equals(
        self,
        current: Any,
        expected: Any,
        msg: str = None,
        with_screen: bool = False,
    ) -> str | None:
        """Check that current does not equal expected."""
        message = (
            f"{base_msg(msg)}\n"
            f"Current - {convert_container(current)}\n"
            f"Expected - {convert_container(expected)}\n"
        )
        self.soft_assert_false(current == expected, message, with_screen)

    def soft_assert_in(
        self,
        member: Any,
        container: Any,
        msg: str = None,
        with_screen: bool = False,
    ) -> str | None:
        """Check that member is in container."""
        msg = (
            f"{base_msg(msg)}\n"
            f"Member: '{member}'\n not found in\n"
            f"Container: '{convert_container(container)}'\n"
        )
        return self.soft_assert_true(member in container, msg, with_screen)

    def soft_assert_not_in(
        self,
        member: Any,
        container: Any,
        msg: str = None,
        with_screen: bool = False,
    ) -> str | None:
        """Check that member is not in container."""
        msg = (
            f"{base_msg(msg)}\n"
            f"Member: '{member}'\n unexpectedly found in\n"
            f"Container: '{convert_container(container)}'\n"
        )
        return self.soft_assert_true(member not in container, msg, with_screen)

    def soft_assert_less_equal(
        self, a: Any, b: Any, msg: str = None, with_screen: bool = False
    ) -> str | None:
        """Check that a is less than or equal to b."""
        error = f"{base_msg(msg)}{a} not less than or equal to {b}\n"
        return self.soft_assert_true(a <= b, error, with_screen)

    def soft_assert_less(
        self, a: Any, b: Any, msg: str = None, with_screen: bool = False
    ) -> str | None:
        """Check that a is less than b."""
        error = f"{base_msg(msg)}{a} not less than {b}\n"
        return self.soft_assert_true(a < b, error, with_screen)

    def soft_assert_greater_equal(
        self, a: Any, b: Any, msg: str = None, with_screen: bool = False
    ) -> str | None:
        """Check that a is greater than or equal to b."""
        error = f"{base_msg(msg)}{a} not greater than or equal to {b}\n"
        return self.soft_assert_true(a >= b, error, with_screen)

    def soft_assert_greater(
        self, a: Any, b: Any, msg: str = None, with_screen: bool = False
    ) -> str | None:
        """Check that a is greater than b."""
        error = f"{base_msg(msg)}{a} not greater than {b}\n"
        return self.soft_assert_true(a > b, error, with_screen)

    def soft_assert_regexp_matches(
        self, text: str, expected_regexp: str, msg: str = None
    ) -> str | None:
        """Fail the test unless the text matches the regular expression."""
        pattern = re.compile(expected_regexp)
        result = pattern.search(text)

        if not result:
            error = (
                f"{base_msg(msg)}"
                f"Regexp didn't match."
                f"Pattern {str(pattern.pattern)} not found in {str(text)}\n"
            )
            self.assertion_errors.append(f"{error}\n{self._callsite_line()}\n")
            return error

    def soft_assert_disable(
        self, element: Any, msg: str = None, with_screen: bool = False
    ) -> str:
        """Check that element has 'disabled' attribute."""
        default_msg = "" if msg else f"Not disabled {element}\n"
        error = f"{base_msg(msg)}{default_msg}"
        return self.soft_assert_true(
            element.get_attribute("disabled"), error, with_screen
        )

    def soft_assert_is_none(self, obj: Any, msg: str = None) -> str:
        """Check that object is None."""
        default_msg = "" if msg else f"{obj} is not None.\n"
        error = f"{base_msg(msg)}{default_msg}"
        return self.soft_assert_true(obj is None, error)

    def soft_assert_is_not_none(self, obj: Any, msg: str = None) -> str:
        """Check that object is not None."""
        default_msg = "" if msg else "Unexpectedly None.\n"
        error = f"{base_msg(msg)}{default_msg}"
        return self.soft_assert_true(obj is not None, error)

    def soft_assert_is_instance(
        self, obj: Any, cls: Any, msg: str = None
    ) -> str:
        """Check that object is instance of given class."""
        default_msg = "" if msg else f"{obj} is not an instance of {cls}.\n"
        error = f"{base_msg(msg)}{default_msg}"
        return self.soft_assert_true(isinstance(obj, cls), error)

    def soft_assert_equals_text_with_ignore_spaces_and_register(
        self,
        current_text: str,
        expected_text: str,
        msg: str = "Invalid checked text.",
        with_screen: bool = False,
    ) -> None:
        """Compare texts ignoring spaces and case."""
        current = get_text_with_ignore_whitespace_symbols(current_text)
        expected = get_text_with_ignore_whitespace_symbols(expected_text)
        if not current:
            msg = "Warning: current text is None!"
        self.soft_assert_equals(
            current.lower(),
            expected.lower(),
            f"{msg}\nCurrent text without formatting: {current_text}"
            f"\nExpected text without formatting: {expected_text}",
            with_screen=with_screen,
        )

    def soft_assert_schemas(
        self, current: dict, expected: dict, msg: str = ""
    ) -> str | None:
        """Validate dictionary schema using json_checker."""
        try:
            Checker(expected, soft=True).validate(current)
        except CheckerError as e:
            error = f"{msg}\n{e}\n"
            self.assertion_errors.append(f"{error}\n{self._callsite_line()}\n")
            return error

    def assert_keys_and_instances(
        self,
        actual_dict: dict,
        expected_dict: dict,
        can_be_null: list = None,
        msg: str = None,
    ) -> None:
        """Check that dict keys and value types match expected."""
        assert actual_dict, "Actual dict is empty, check your data"

        self.soft_assert_equals(
            sorted(iter(actual_dict.keys())),
            sorted(iter(expected_dict.keys())),
            "Wrong keys list.",
        )
        for actual_key, actual_value in actual_dict.items():
            self.soft_assert_in(
                member=actual_key,
                container=expected_dict,
                msg=f'Not expected key "{actual_key}".',
            )
            if actual_key in expected_dict:
                expected_value = (
                    type(None)
                    if actual_value is None and actual_key in (can_be_null or [])
                    else expected_dict[actual_key]
                )
                message = f"({msg})" if msg else ""
                self.soft_assert_true(
                    expr=isinstance(actual_value, expected_value),
                    msg=(
                        f"Wrong object instance class.\n"
                        f'Key "{actual_key}" value is "{type(actual_value)}", '
                        f'expected "{expected_value}". {message}'
                    ),
                )

    def soft_assert_dicts_with_ignore_types(
        self, current: Any, expected: Any, msg: str = ""
    ) -> None:
        """Compare two dicts ignoring value types."""
        expected_dict = _check_namedtuple(expected)
        current_dict = _check_namedtuple(current)
        errors = ""
        for actual_key, actual_value in current_dict.items():
            if actual_key in expected_dict:
                if str(actual_value) != str(expected_dict[actual_key]):
                    errors += (
                        f"\nKey {actual_key} "
                        f"\nHas incorrect value: {actual_value}  "
                        f"Expected: {expected_dict[actual_key]} \n"
                    )
        if errors:
            self.assertion_errors.append(
                f"{msg}{errors}"
                f"\nExpected dict: {convert_container(expected_dict)}"
                f"\nCurrent dict: {convert_container(current_dict)}"
                f"\n{self._callsite_line()}\n"
            )

    def soft_assert_equal_dict_from_clerk_magic(
        self,
        current_dict: dict,
        expected_dict: dict,
        msg: str = None,
        coerce: CoerceDict = None,
    ) -> None:
        """Compare dicts by type, coercing values if needed."""

        def _equal(
            field_name: str, current: str, expected: str, type_coerce: CoerceDict
        ) -> bool:
            if type_coerce is not None and field_name in type_coerce:
                cast = type_coerce[field_name]
                return cast(current) == cast(expected)
            if type(current) is type(expected):
                return current == expected
            return str(current) == str(expected)

        missing = []
        unequal = []
        for key, value in expected_dict.items():
            if key not in current_dict:
                missing.append((key, value))
            elif not _equal(key, value, current_dict[key], coerce):
                unequal.append((
                    key,
                    f"\n\t\tproduct_data - '{value}'"
                    f" != analytic_data - '{current_dict[key]}'",
                ))

        unequal_check = "\n\t".join([":".join(map(str, a)) for a in unequal])
        missing_check = "\n\t".join([":".join(map(str, a)) for a in missing])
        self.soft_assert_true(
            (missing, unequal) == ([], []),
            msg=(
                f"{base_msg(msg)}\n"
                f"Unequal_keys:\n\t{unequal_check}\n"
                f"\nMissed_keys:\n\t[{missing_check}]"
            ),
        )


class RequestSoftAssertion(BaseSoftAssertion):
    """Soft assertions for request-based checks."""

    @property
    def url(self) -> str:
        return self.session.url

    def base_msg(self, msg: str = None) -> str:
        """Return formatted message including URL and optional extra message."""
        url = f"\n{self.url}" if self.url else ""
        exception = f"\n{msg}" if msg else ""
        return f"{url}{exception}"


class WebDriverSoftAssertion(BaseSoftAssertion):
    """Soft assertions for WebDriver-based checks."""

    @property
    def driver(self) -> WebDriver:
        return self.driver

    @property
    def url(self) -> str:
        return self.driver.current_url

    def base_msg(self, default_msg: str = None, extend_msg: str = None) -> str:
        """Return formatted message including default, extended, and URL."""
        default = f"Default message: {default_msg}\n" if default_msg else ""
        extend = f"Extend message: {extend_msg}\n" if extend_msg else ""
        return f"{default}{extend}\nURL: {self.url}"

    def soft_assert_page_title(
        self, expected_title: str, msg: str = None
    ) -> None:
        """Check that page title matches expected."""
        default_msg = "Wrong page title."
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_equals(self.driver.title, expected_title, error)

    def soft_assert_current_url(
        self, expected_url: str, msg: str = None
    ) -> None:
        """Check that current URL matches expected."""
        default_msg = "Wrong current url."
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_equals(self.url, expected_url, error)

    def soft_assert_current_url_contains(
        self, url_mask: str, msg: str = None
    ) -> None:
        """Check that current URL contains mask."""
        default_msg = f"URL {str(self.url)} doesn't contain {str(url_mask)}."
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_in(url_mask, self.url, error)

    def soft_assert_current_url_not_contains(
        self, url_mask: str, msg: str = None
    ) -> None:
        """Check that current URL does not contain mask."""
        default_msg = f"URL {str(self.url)} contains {str(url_mask)}."
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_not_in(url_mask, self.url, error)

    def soft_assert_element_is_present(
        self, element: Any, msg: str = None
    ) -> None:
        """Check that element is present in DOM."""
        default_msg = (
            f"Element {element.by}={element.locator} is not present"
            f" on page at current time."
        )
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_true(element.is_present(), error, with_screen=True)

    def soft_assert_element_is_not_present(
        self, element: Any, msg: str = None
    ) -> None:
        """Check that element is not present in DOM."""
        default_msg = f"Element {element.by}={element.locator} is found on page."
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_false(element.is_present(), error, with_screen=True)

    def soft_assert_element_is_displayed(
        self, element: Any, msg: str = None
    ) -> None:
        """Check that element is visible to user."""
        default_msg = (
            f"Element {element.by}={element.locator} is not visible to a user."
        )
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_true(element.is_displayed(), error, with_screen=True)

    def soft_assert_element_is_not_displayed(
        self, element: Any, msg: str = None
    ) -> None:
        """Check that element is not visible to user."""
        default_msg = (
            f"Element {element.by}={element.locator} is visible to a user."
        )
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_false(element.is_displayed(), error, with_screen=True)

    def soft_assert_element_displayed_in_viewport(
        self, element: Any, msg: str = None
    ) -> None:
        """Check that element is inside viewport."""
        element_in_viewport = self.driver.execute_script(
            """
            function elementInViewport(el) {
                var top = el.offsetTop;
                var left = el.offsetLeft;
                var width = el.offsetWidth;
                var height = el.offsetHeight;

                while(el.offsetParent) {
                  el = el.offsetParent;
                  top += el.offsetTop;
                  left += el.offsetLeft;
                }
                return (
                    top >= window.pageYOffset &&
                    left >= window.pageXOffset &&
                    (top + height) <= (
                        window.pageYOffset + window.innerHeight
                    ) &&
                    (left + width) <= (
                        window.pageXOffset + window.innerWidth
                    )
                );
            }
            element = arguments[0];
            return elementInViewport(element);
            """,
            element.lookup(),
        )

        if element_in_viewport is not True:
            default_msg = (
                f"Element {element.by}={element.locator}"
                f" not displayed in viewport"
            )
            error = (
                f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}\n\n"
            )
            self.assertion_errors.append(f"{error}\n{self._callsite_line()}\n")

    def soft_assert_image_status_code(
        self, image: Any, status_code: int = 200, msg: str = None
    ) -> str | None:
        """Check that image URL returns expected status code."""
        if not image.get_attribute("src"):
            return self.assertion_errors.append(
                f"{base_msg(msg)}"
                "Image does not have attribute"
                f" 'src'\n\n{self._callsite_line()}\n"
            )

        img_url = image.get_attribute("src")
        response = http.request(url=img_url, method="GET")

        assert_msg = msg if msg else "img status code != 200"
        self.soft_assert_equals(
            response.status,
            status_code,
            msg=assert_msg,
        )
